import asyncio  # noqa: F401
import json
from pathlib import Path
from pprint import pprint

import click

from candore.errors import ModeError
from candore.modules.api_lister import APILister
from candore.modules.comparator import Comparator
from candore.modules.extractor import Extractor
from candore.modules.finder import Finder
from candore.modules.report import Reporting


class Candore:
    def __init__(self, settings):
        self.settings = settings
        self.api_lister = APILister(settings=self.settings)

    def list_endpoints(self):
        return self.api_lister.lister_endpoints()

    async def save_all_entities(self, mode, output_file, full):
        """Save all the entities to a json file

        :param mode: Pre or Post
        :param output_file: Output file name
        :param full: If True, save entities from all pages of the components,
            else just saves first page
        :return: None
        """
        if mode not in ["pre", "post"]:
            raise ModeError("Extracting mode must be 'pre' or 'post'")

        async with Extractor(settings=self.settings, apilister=self.api_lister) as extractor:
            if full:
                extractor.full = True
            data = await extractor.extract_all_entities()

        if not data:
            click.echo("Entities data is not data found!")

        file_path = Path(output_file) if output_file else Path(f"{mode}_entities.json")
        with file_path.open(mode="w") as entfile:
            json.dump(data, entfile)
        click.echo(f"Entities data saved to {file_path}")

    def compare_entities(
        self,
        pre_file=None,
        post_file=None,
        inverse=None,
        output=None,
        report_type=None,
        record_evs=None,
    ):
        comp = Comparator(settings=self.settings)
        if record_evs:
            comp.record_evs = True
        results = comp.compare_json(pre_file=pre_file, post_file=post_file, inverse=inverse)
        reporter = Reporting(results=results)
        reporter.generate_report(output_file=output, output_type=report_type, inverse=inverse)

    def find_path(self, path, json_file, delimiter):
        finder = Finder()
        data = finder.find(path=path, json_file=json_file, delimiter=delimiter)
        if data:
            pprint(data)



# -*- encoding: utf-8 -*-
"""
Provides a class with methods that parse the api correctly.

Parser classes must currently implement the following methods:
    pull_links - Returns a list of links to visit from the API's base url.
    yaml_format - Returns yaml-friendly dict of the compiled data.
    scrape_content - Returns a dict of params and paths from a single page.
"""
import attr
from logzero import logger


@attr.s()
class APIPie:
    """Parser class for Ruby's APIPie apidoc generator"""

    _data = attr.ib(default=attr.Factory(dict), repr=False)
    params = attr.ib(default=attr.Factory(dict), repr=False)

    @staticmethod
    def _compile_method(method_dict):
        """form the parameters and paths lists"""
        params = [
            (
                f'{param["name"]} ~ '
                f'{"required" if param["required"] else "optional"} ~ '
                f'{param["expected_type"]}'
            )
            for param in method_dict["params"]
        ]
        paths = [f'{path["http_method"].upper()} {path["api_url"]}' for path in method_dict["apis"]]
        return {"paths": paths, "params": params}

    def scrape_content(self, result):
        """Compile the data into their corresponding classifications"""
        entity_docs = result.json()["docs"]["resources"]
        for name, data in entity_docs.items():
            logger.debug(f"Compiling {name} with {len(data['methods'])} methods")
            self._data[name] = {"methods": []}
            for method in data["methods"]:
                self._data[name]["methods"].append({method["name"]: self._compile_method(method)})
                self.params.update({param["name"]: param for param in method["params"]})

    def yaml_format(self, ingore=None):
        """Return the compiled data in a yaml-friendly format"""
        return self._data



import json

from candore.modules.variations import Variations, Constants
from candore.utils import last_index_of_element, is_list_contains_dict


class Comparator:
    def __init__(self, settings):
        self.big_key = []
        self.big_diff = {}
        self.big_constant = {}
        self.record_evs = False
        self.variations = Variations(settings)
        self.constants = Constants(settings)
        self.expected_variations = self.variations.expected_variations
        self.skipped_variations = self.variations.skipped_variations
        self.expected_constants = self.constants.expected_constants
        self.skipped_constants = self.constants.skipped_constants



    def remove_verifed_key(self, key):
        reversed_bk = self.big_key[::-1]
        if key in reversed_bk:
            reversed_bk.remove(key)
        self.big_key = reversed_bk[::-1]

    def remove_path(self, identy):
        id_index = last_index_of_element(self.big_key, identy)
        if id_index == -1:
            return
        self.big_key = self.big_key[:id_index]

    def record_variation(self, pre, post, var_details=None):
        big_key = [str(itm) for itm in self.big_key]
        full_path = "/".join(big_key)
        var_full_path = "/".join([itm for itm in self.big_key if not isinstance(itm, int)])
        if var_full_path in self.expected_variations or var_full_path in self.skipped_variations:
            if self.record_evs:
                variation = {
                    "pre": pre,
                    "post": post,
                    "variation": var_details or "Expected(A)",
                }
                self.big_diff.update({full_path: variation})
        elif (
            var_full_path not in self.expected_variations
            and var_full_path not in self.skipped_variations
        ):
            variation = {"pre": pre, "post": post, "variation": var_details or ""}
            self.big_diff.update({full_path: variation})

    def record_constants(self, pre, post, var_details=None):
        big_key = [str(itm) for itm in self.big_key]
        full_path = "/".join(big_key)
        var_full_path = "/".join([itm for itm in self.big_key if not isinstance(itm, int)])
        if var_full_path in self.expected_constants or var_full_path in self.skipped_constants:
            if self.record_evs:
                variation = {
                    "pre": pre,
                    "post": post,
                    "constant": var_details or "Expected(A)",
                }
                self.big_constant.update({full_path: variation})
        elif (
            var_full_path not in self.expected_constants
            and var_full_path not in self.skipped_constants
        ):
            variation = {"pre": pre, "post": post, "constant": var_details or ""}
            self.big_constant.update({full_path: variation})

    def _is_data_type_dict(self, pre, post, unique_key=""):
        if (pre and 'id' in pre) and (post and 'id' in post):
            # Dont compare the entities if the ids are not the same
            if pre['id'] != post['id']:
                return
        for pre_key in pre:
            if pre_key in post:
                key = pre_key
                self.compare_all_pres_with_posts(pre[key], post[key], unique_key=key)
            else:
                self.compare_all_pres_with_posts(
                    pre[pre_key],
                    None,
                    unique_key=pre_key,
                    var_details="Post lookup key missing",
                )
        self.remove_path(unique_key)

    def _is_data_type_list_contains_dict(self, pre, post):
        for pre_entity in pre:
            if not pre_entity:
                continue
            for post_entity in post:
                if not post_entity:
                    continue
                if "id" in pre_entity:
                    if pre_entity["id"] == post_entity["id"]:
                        self.compare_all_pres_with_posts(
                            pre_entity, post_entity, unique_key=pre_entity["id"]
                        )
                else:
                    key = list(pre_entity.keys())[0]
                    if pre_entity[key] == post_entity[key]:
                        self.compare_all_pres_with_posts(
                            pre_entity[key], post_entity[key], unique_key=key
                        )
            if "id" in pre_entity:
                self.remove_path(pre_entity["id"])
            else:
                self.remove_path(pre_entity[list(pre_entity.keys())[0]])

    def _is_data_type_list(self, pre, post, unique_key=""):

        def custom_key(elem):
            return 'None' if elem is None else str(elem)

        if not is_list_contains_dict(pre):
            if sorted(pre, key=custom_key) != sorted(post, key=custom_key):
                self.record_variation(pre, post)
            else:
                self.record_constants(pre, post)
        else:
            self._is_data_type_list_contains_dict(pre, post)
        self.remove_path(unique_key)

    def compare_all_pres_with_posts(self, pre_data, post_data, unique_key="", var_details=None):
        if unique_key:
            self.big_key.append(unique_key)
        if type(pre_data) is dict:
            self._is_data_type_dict(pre_data, post_data, unique_key=unique_key)
        elif type(pre_data) is list:
            self._is_data_type_list(pre_data, post_data, unique_key=unique_key)
        else:
            if pre_data != post_data:
                self.record_variation(pre_data, post_data, var_details)
            else:
                self.record_constants(pre_data, post_data, var_details)
            self.remove_verifed_key(unique_key)

    def compare_json(self, pre_file, post_file, inverse):
        pre_data = post_data = None

        with open(pre_file, "r") as fpre:
            pre_data = json.load(fpre)

        with open(post_file, "r") as fpost:
            post_data = json.load(fpost)

        self.compare_all_pres_with_posts(pre_data, post_data)
        if not inverse:
            return self.big_diff
        else:
            return self.big_constant



# -*- encoding: utf-8 -*-
"""Determine the changes between two API versions."""
from pathlib import Path

import attr
import yaml
from logzero import logger

from apix.helpers import get_latest
from apix.helpers import get_previous
from apix.helpers import load_api


@attr.s()
class VersionDiff:
    api_name = attr.ib(default=None)
    ver1 = attr.ib(default=None)
    ver2 = attr.ib(default=None)
    data_dir = attr.ib(default=None)
    compact = attr.ib(default=False)
    mock = attr.ib(default=False, repr=False)
    _vdiff = attr.ib(default={})

    def __attrs_post_init__(self):
        """Load the API versions, if not provided"""
        if not self.api_name:
            self.api_name = get_latest(data_dir=self.data_dir, mock=self.mock)
        if not self.ver1:
            # get the latest saved version
            self.ver1 = get_latest(api_name=self.api_name, data_dir=self.data_dir, mock=self.mock)
        if not self.ver2:
            # get the version before ver1
            self.ver2 = get_previous(self.api_name, self.ver1, self.data_dir, self.mock)

    @staticmethod
    def _truncate(diff_dict):
        """Strip all extra information from a diff"""
        compact_diff = {}
        for parent, children in diff_dict.items():
            compact_diff[parent] = []
            for meth in children["methods"]:
                compact_diff[parent].append(list(meth.keys())[0])
        return compact_diff

    def _dict_diff(self, dict1, dict2):
        """Recursively search a dictionary for differences"""
        added, changed = {}, {}
        if dict1 == dict2:
            return added, changed
        for key, values in dict1.items():
            if key in dict2:
                if not values == dict2[key]:
                    if isinstance(values, dict):
                        res, chng = self._dict_diff(values, dict2[key])
                        if res:
                            added[key] = res
                        if chng:
                            changed[key] = chng
                    elif isinstance(values, list):
                        res, chng = self._list_diff(values, dict2[key])
                        if res:
                            added[key] = res
                        if chng:
                            changed[key] = chng
                    else:
                        logger.debug(f"Adding {key} => {values}")
                        added[key] = values
            else:
                logger.debug(f"Adding {key} => {values}")
                added[key] = values
        return added, changed

    def _list_diff(self, list1, list2):
        """Recursively search a list for differences"""
        added, changed = [], []
        if list1 == list2:
            return added, changed
        for item in list1:
            if isinstance(item, dict):
                found = False
                for key in item:
                    for i, needle in enumerate(list2):
                        if key in needle:
                            found = True
                            res, chng = self._dict_diff(item, list2[i])
                            if res:
                                added.append(res)
                            if chng:
                                changed.append(chng)
                if not found:
                    logger.debug(f"Adding {item}")
                    added.append(item)
            elif isinstance(item, list):
                res, chng = self._list_diff(item, list2[list2.index(item)])
                if res:
                    added.append(res)
                if chng:
                    changed.append(chng)
            elif " ~ " in item:
                name = item.split("~")[0].strip()
                found = False
                for needle in list2:
                    # look for matching names
                    if name == needle.split("~")[0].strip():
                        found = True
                        if item != needle:
                            logger.debug(f"{item} changed to {needle}")
                            changed.append(item)
                if not found:
                    logger.debug(f"Adding {item}")
                    added.append(item)
            else:
                if item not in list2:
                    logger.debug(f"Adding {item}")
                    added.append(item)
        return added, changed

    def diff(self):
        """Determine the diff between ver1 and ver2"""
        if not self.ver1:
            logger.warning("No ver1 API found.")
            return None
        if not self.ver2:
            logger.warning("No ver2 API found.")
            return None
        logger.info(f"Performing diff between {self.ver1} and {self.ver2}")

        ver1_content = load_api(self.api_name, self.ver1, self.data_dir, self.mock)
        logger.debug(f"Loaded {self.ver1}.")
        ver2_content = load_api(self.api_name, self.ver2, self.data_dir, self.mock)
        logger.debug(f"Loaded {self.ver2}.")

        added, changed = self._dict_diff(ver1_content, ver2_content)
        logger.debug("Determined added/changed content.")
        removed, _ = self._dict_diff(ver2_content, ver1_content)
        logger.debug("Determined removed content.")
        if self.compact:
            added = VersionDiff._truncate(added)
            changed = VersionDiff._truncate(changed)
            removed = VersionDiff._truncate(removed)
        self._vdiff = {
            f"Added in {self.ver1} since {self.ver2}": added,
            f"Changed in {self.ver1} since {self.ver2}": changed,
            f"Removed since {self.ver2}": removed,
        }

    def save_diff(self, return_path=False):
        """Save the currently stored diff"""
        if not self._vdiff:
            logger.warning("No data to be saved. Exiting.")
            return

        if self.mock:
            fpath = Path(
                f"{self.data_dir}tests/APIs/{self.api_name}"
                f"/{self.ver2}-to-{self.ver1}-diff.yaml"
            )
        else:
            ftype = "comp-diff" if self.compact else "diff"
            fpath = Path(
                f"{self.data_dir}APIs/{self.api_name}/" f"{self.ver2}-to-{self.ver1}-{ftype}.yaml"
            )
        if fpath.exists():
            fpath.unlink()
        # create the directory, if it doesn't exist
        fpath.parent.mkdir(parents=True, exist_ok=True)
        fpath.touch()
        logger.info(f"Saving results to {fpath}")
        with fpath.open("w+") as outfile:
            yaml.dump(self._vdiff, outfile, default_flow_style=False)
        if return_path:
            return fpath



import asyncio
from pprint import pprint

import click

from candore import Candore
from candore.config import candore_settings

# Click Interactive for Cloud Resources Cleanup


@click.group(
    help="A data integrity validation CLI tool for upgradable products",
    invoke_without_command=True,
)
@click.option("--version", is_flag=True, help="Installed version of candore")
@click.option("--settings-file", "-s", default=None, help="Settings file path")
@click.option("--components-file", "-c", default=None, help="Components file path")
@click.pass_context
def candore(ctx, version, settings_file, components_file):
    if version:
        import pkg_resources

        ver = pkg_resources.get_distribution("candore").version
        click.echo(f"Version: {ver}")
    candore_obj = Candore(
        settings=candore_settings(
            option_settings_file=settings_file, option_components_file=components_file
        )
    )
    ctx.__dict__["candore"] = candore_obj


@candore.command(help="List API lister endpoints from Product")
@click.pass_context
def apis(ctx):
    """List API lister endpoints from Product"""
    print("List of API lister endpoints from Product")
    candore_obj = ctx.parent.candore
    pprint(candore_obj.list_endpoints())


@candore.command(help="Extract and save data using API lister endpoints")
@click.option("--mode", type=str, help="The mode must be 'pre' or 'post'")
@click.option("-o", "--output", type=str, help="The output file name")
@click.option("--full", is_flag=True, help="Extract data from all the pages of a component")
@click.pass_context
def extract(ctx, mode, output, full):
    loop = asyncio.get_event_loop()
    candore_obj = ctx.parent.candore
    loop.run_until_complete(candore_obj.save_all_entities(mode=mode, output_file=output, full=full))


@candore.command(help="Compare pre and post upgrade data")
@click.option("--pre", type=str, help="The pre upgrade json file")
@click.option("--post", type=str, help="The post upgrade json file")
@click.option("-i", "--inverse", is_flag=True, help="Inverse comparison, shows whats not changed")
@click.option("-o", "--output", type=str, help="The output file name")
@click.option(
    "-t",
    "--report-type",
    type=str,
    default="json",
    help="The type of report GSheet, JSON, or webpage",
)
@click.option("--record-evs", is_flag=True, help="Record Expected Variations in reporting")
@click.pass_context
def compare(ctx, pre, post, inverse, output, report_type, record_evs):
    candore_obj = ctx.parent.candore
    candore_obj.compare_entities(
        pre_file=pre,
        post_file=post,
        inverse=inverse,
        output=output,
        report_type=report_type,
        record_evs=record_evs,
    )


@candore.command(help="JSON Reader for reading the specific path data from entities data file")
@click.option(
    "--path",
    type=str,
    help="The path to search the data from.\n"
    "Path contents could divided by some delimiter.\n"
    "e.g entity/5/description",
)
@click.option(
    "--data-file", type=str, help="The data file from which to search the data on a given path"
)
@click.option("--delimiter", type=str, default='/', help="Settings file path. Default is '/'")
@click.pass_context
def reader(ctx, path, data_file, delimiter):
    candore_obj = ctx.parent.candore
    candore_obj.find_path(path=path, json_file=data_file, delimiter=delimiter)


if __name__ == "__main__":
    candore()



"""
An utility helpers module
"""
from pathlib import Path
import yaml


def last_index_of_element(arr, element):
    for i in range(len(arr) - 1, -1, -1):
        if arr[i] == element:
            return i
    return -1


def is_list_contains_dict(_list):
    contains_dict = any(isinstance(element, dict) for element in _list)
    return contains_dict


def yaml_reader(file_path=None):
    templates_path = Path(file_path)
    if not templates_path.exists():
        print(f"The file {templates_path} does not exist.")
    with templates_path.open() as yaml_file:
        yaml_data = yaml.safe_load(yaml_file)
    return yaml_data


def get_yaml_paths(yaml_data, prefix="", separator="/"):
    paths = []
    if isinstance(yaml_data, dict):
        for key, value in yaml_data.items():
            paths.extend(get_yaml_paths(value, f"{prefix}{key}{separator}"))
    elif isinstance(yaml_data, list):
        for item in yaml_data:
            paths.extend(get_yaml_paths(item, prefix, separator))
    else:
        paths.append(f"{prefix}{yaml_data}")
    return paths


"""Centralized Reporting utilities for compared results"""
import csv
import json
from pathlib import Path

# from .webapp import display_json_table, render_webpage


class Reporting:
    """Centralized Reporting utilities for compared results"""

    def __init__(self, results):
        """Initialize the Reporting class

        Args:
            comparator (Comparator): The comparator object to use for reporting
        """
        self.results = results

    def generate_report(self, output_file, output_type, inverse):
        """Generate a report of the compared results

        Args:
            output_file (str): The file to write the report to
            output_type (str): The type of report to generate json / CSV
            inverse (bool): Shows what not changed in comparison results
        Returns:
            None
        Raises:
            ValueError: If the output_type is not supported
        """
        if output_type == "json":
            self._generate_json_report(output_file)
        elif output_type == "html":
            print('The HTML reporting is not implemented yet! Try next time!')
        elif output_type == "csv":
            self._generate_csv_report(output_file, inverse=inverse)
        else:
            raise ValueError("Output type {} not supported".format(output_type))

    def _generate_json_report(self, output_file):
        """Generate a JSON report of the compared results

        Args:
            output_file (str): The file to write the report to
        Returns:
            None
        """
        if not output_file:
            output_file = "results.json"
        output_file = Path(output_file)
        # Write the JSON report to the output file
        output_file.write_text(json.dumps(self.results, indent=4))
        print("Wrote JSON report to {}".format(output_file))
        print("JSON report contains {} results".format(len(self.results)))

    def _generate_html_report(self):
        """Generate an HTML report of the compared results

        Args:
            output_file (str): The file to write the report to
        Returns:
            None
        """
        # display_json_table(results_json=self.results)
        # render_webpage()
        print("HTML report is ready to view at: http://localhost:5000")

    def _generate_csv_report(self, output_file, inverse):
        """Generate a CSV report of the compared results

        Args:
            output_file (str): The file to write the report to
        Returns:
            None
        """
        if not output_file:
            output_file = "results.csv"
        output_file = Path(output_file)
        # Convert json to csv and write to output file
        csv_writer = csv.writer(output_file.open("w"))
        # Table Column Names
        columns = ["Path", "Pre-Upgrade", "Post-Upgrade", "Variation?" if not inverse else 'Constant?']
        csv_writer.writerow(columns)
        # Writing Rows
        for var_path, vals in self.results.items():
            csv_writer.writerow([
                var_path, vals["pre"], vals["post"],
                vals["variation" if not inverse else "constant"]])
        print("Wrote CSV report to {}".format(output_file))
        print("CSV report contains {} results".format(len(self.results)))



"""
A module responsible for calculating expected and skipped variations from
`conf/variations` yaml file and convert them into processable list
"""
from functools import cached_property
from candore.utils import yaml_reader, get_yaml_paths

import yaml


class Variations:
    def __init__(self, settings):
        self.settings = settings

    @cached_property
    def variations(self):
        yaml_data = yaml_reader(file_path=self.settings.candore.var_file)
        return yaml_data

    @cached_property
    def expected_variations(self):
        return get_yaml_paths(yaml_data=self.variations.get("expected_variations"))

    @cached_property
    def skipped_variations(self):
        return get_yaml_paths(yaml_data=self.variations.get("skipped_variations"))


class Constants:
    def __init__(self, settings):
        self.settings = settings

    @cached_property
    def constants(self):
        yaml_data = yaml_reader(file_path=self.settings.candore.constant_file)
        return yaml_data

    @cached_property
    def expected_constants(self):
        return get_yaml_paths(yaml_data=self.constants.get("expected_constants"))

    @cached_property
    def skipped_constants(self):
        return get_yaml_paths(yaml_data=self.constants.get("skipped_constants"))



# -*- encoding: utf-8 -*-
"""
Provides a class with methods that parse the api correctly.

Parser classes must currently implement the following methods:
    pull_links - Returns a list of links to visit from the API's base url.
    yaml_format - Returns yaml-friendly dict of the compiled data.
    scrape_content - Returns a dict of params and paths from a single page.

https://www.google.com/search?q=apix
"""
import attr
from lxml import html


@attr.s()
class TestParser:
    """Parser class for testing purposes only."""

    _data = attr.ib(default={}, repr=False)

    def _data_to_yaml(self, index):
        """Translate a url and text into 'paths' and 'parameters'"""
        url = index
        split_url = url.split(".")
        name = split_url[-1]
        first = self._data[index][0]
        rest = " ~ ".join(self._data[index][1:])
        return url, {name: {"paths": first, "parameters": rest}}

    def yaml_format(self, data):
        """compile all data into a yaml-compatible dict"""
        self._data, yaml_data = data, {}
        for index in self._data:
            ent, res = self._data_to_yaml(index)
            if ent and not yaml_data.get(ent, None):
                yaml_data[ent] = {"content": [res]}
            elif ent:
                yaml_data[ent]["content"].append(res)
        return yaml_data

    @staticmethod
    def pull_links(result, base_path):
        """return all desired links from the target page"""
        g_links = html.fromstring(result.content).iterlinks()
        links, last = [], None
        for link in g_links:
            url = link[2].replace("../", "")
            if "JacobCallahan" in url and "sparkline" not in url and link[0].text and url != last:
                links.append((link[0].text, url))
                last = url
        return links

    @staticmethod
    def scrape_content(content):
        """take the title text from a page, if it exists"""
        tree = html.fromstring(content)
        title = tree.xpath("//head/title")
        if title:
            title = title[0].text
        else:
            title = "This page had no title for some reason"
        return [x.strip() for x in title.split() if x]
